import pickle
import os, sys
import random
# import setuptools
import shutil
import cv2
from os.path import join, exists, dirname
from pprint import pprint as pp
# print(os.getcwd()) # Thư mục console đang chạy
# print(__file__) # Thư mục file code
codeDir=dirname(os.path.abspath(__file__))
sys.path.append(codeDir)

# pp(sys.path)

from ParamsBase import tactParametters
from tqdm import tqdm
from yolo_boxes import yolo_str_to_xyxy, xyxyn_to_xyxy
from PIL import Image


APP_NAME='Image_Augmentation'

class Parameters(tactParametters):
    def __init__(self, ModuleName="TACT"):
        super().__init__(saveParam_onlyThis_APP_NAME=True)
        self.AppName = APP_NAME
        self.Ready_to_run = False # Nếu bắt buộc phải config thì đặt cái này = False, khi nào user chỉnh sang True thì mới cho chạy
        self.HD = {
            "Brightness": "Adjust brightness (value: -255 to 255)",
            "Shadow_Brightness": "Adjust Shadow_Brightness (value: -255 to 255)",
            "Contrast": "Adjust contrast (alpha > 1 increases contrast, 0 < alpha < 1 decreases)",
            "Saturation": "Convert to HSV, adjust saturation, then back to BGR",
            "Hue": "Convert to HSV, adjust hue, then back to BGR",
            "cover_yolo_string": "'0 0.496094 0.772135 0.527344 0.453125' (ví dụ xâu), cái này nếu có giá trị, nó sẽ crop ảnh theo tọa độ trong này, TODO: tính lại tọa độ label theo cái này",
            "cover_yolo_string__resize_after_crop_HW": "Là 1 list, nếu không muốn resize, để thành list rỗng: []",
            "Max_random_image_to_Aug=x": "x==0: Aug tất cả các ảnh, x>0: chỉ Aug random x ảnh thôi, để test trong quá trình dò tim các tham số",
            "Cách dùng 1": "lệnh chạytrong CMD: 'ntanh_aug', khi hiển thị ảnh: bấm SPACE để tạm dừng/chạy tiếp, bấm ESC để thoát",
            "Cách dùng 2": "Muốn đưa tham số nào về mặc định thì xóa nó đi rồi chạy lại",
        }
        self.Intro="Chương trình này dành riêng cho việc augmentation ảnh nhằm tạo ra ảnh có các kiểu khác nhau phục vụ cho mục đích training model"

        self.cover_yolo_string = ""
        self.cover_yolo_string__expand_n_pixel = 20
        self.cover_yolo_string__resize_after_crop_WH=[]
        self.Brightness = 50
        self.Shadow_Brightness=0
        self.Contrast = 1.2
        self.Saturation = 0
        self.Hue = 0
        self.image_folder__input = ""
        self.image_folder_output = ""
        self.Copy_label_when_save_augment_image=True

        self.Display_image=True
        self.Display_image_Stop_to_View_in___cv2_waitkey__ms=1000
        self.Display_image_max_height=800
        self.Max_random_image_to_Aug=10
        self.load_then_save_to_yaml(file_path=f"{APP_NAME}.yml", ModuleName=ModuleName)


mParams = Parameters("Augmentation_images")

def taImshow(title="image",image=None, wait=0):
    H,W=image.shape[:2]
    if H > mParams.Display_image_max_height:
        fxy = mParams.Display_image_max_height / H
        image = cv2.resize(image, None, fx=fxy, fy=fxy)
    cv2.imshow(title, image)
    if wait is not None:
        cv2.waitKey(wait)


def update_labels(
    label_path, x_min_crop, y_min_crop, new_imW, new_imH, old_imW, old_imH
):
    # Cập nhật file nhãn dựa trên ảnh đã crop
    new_labels = []
    with open(label_path, "r") as file:
        lines = file.readlines()
        for line in lines:
            # print(line)
            obj_class, x_center, y_center, width, height = map(float, line.split())

            # Tính tọa độ gốc của bounding box (theo ảnh ban đầu)
            old_x_min = (x_center - width / 2) * old_imW
            old_y_min = (y_center - height / 2) * old_imH
            old_x_max = (x_center + width / 2) * old_imW
            old_y_max = (y_center + height / 2) * old_imH

            # Điều chỉnh bounding box theo vùng đã crop
            new_x_min = max(0, old_x_min - x_min_crop)
            new_y_min = max(0, old_y_min - y_min_crop)
            new_x_max = min(new_imW, old_x_max - x_min_crop)
            new_y_max = min(new_imH, old_y_max - y_min_crop)

            # Bỏ qua các đối tượng không nằm trong vùng crop
            # if new_x_min < new_x_max and new_y_min < new_y_max:
            if 1:
                new_x_center = (new_x_min + new_x_max) / 2 / new_imW
                new_y_center = (new_y_min + new_y_max) / 2 / new_imH
                new_width = (new_x_max - new_x_min) / new_imW
                new_height = (new_y_max - new_y_min) / new_imH

                # Kiểm tra nếu tọa độ trung tâm nằm trong vùng hợp lệ
                # if 0 <= new_x_center <= 1 and 0 <= new_y_center <= 1:
                new_labels.append(
                    f"{int(obj_class)} {new_x_center:0.6f} {new_y_center:0.6f} {new_width:0.6f} {new_height:0.6f}"
                )

    s = "\n".join(new_labels)
    return s


class ImageAugmentation:
    def __init__(self):
        pass

    def Change_Brightness(self, image, value):
        # Adjust brightness
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        h, s, v = cv2.split(hsv)

        # Ensure brightness value is in the range of [-255, 255]
        if value > 0:
            lim = 255 - value
            v[v > lim] = 255
            v[v <= lim] += value
        else:
            lim = -value
            v[v < lim] = 0
            v[v >= lim] += value

        final_hsv = cv2.merge((h, s, v))
        result = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2BGR) 
        return result

    def Change_Shadow_Brightness(self, image, value):
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        h, s, v = cv2.split(hsv)

        # Create a mask for shadows (dark areas)
        shadow_mask = v < 128

        # Adjust brightness in shadow areas
        if value > 0:
            lim = 255 - value
            v[shadow_mask & (v > lim)] = 255
            v[shadow_mask & (v <= lim)] += value
        else:
            lim = -value
            v[shadow_mask & (v < lim)] = 0
            v[shadow_mask & (v >= lim)] += value

        final_hsv = cv2.merge((h, s, v))
        result = cv2.cvtColor(final_hsv, cv2.COLOR_HSV2BGR)

        return result

    def Change_Contrast(self, image, value):
        # Adjust contrast (alpha > 1 increases contrast, 0 < alpha < 1 decreases)
        return cv2.convertScaleAbs(image, alpha=value, beta=0)

    def Change_Saturation(self, image, value):
        # Convert to HSV, adjust saturation, then back to BGR
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        hsv[:, :, 1] = cv2.add(hsv[:, :, 1], value)
        return cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)

    def Change_Hue(self, image, value):
        # Convert to HSV, adjust hue, then back to BGR
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        hsv[:, :, 0] = cv2.add(hsv[:, :, 0], value)
        return cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)

    def Change_Vibrance(self, image, value):
        # Custom vibrance logic, placeholder here
        return image  # Vibrance adjustment not implemented for simplicity
    def restore_coordinates(self, img, restore_options): 
        # Các thông tin từ input
        org_image_hw = restore_options["orgImageHW"]  # (org_height, org_width)
        crop_image = restore_options["cropImage_xyxy"]     # (crop_x1, crop_y1, crop_x2, crop_y2)
        resize_image = restore_options["resizeImage"] # (resize_width, resize_height)
        xyxy = restore_options["xyxy"]              # bbox normalized coords in resizeImage (x_min, y_min, x_max, y_max)
        # xyxy = restore_options["xyxy"]              # bbox normalized coords in resizeImage (x_min, y_min, x_max, y_max)
        # draw_rectangles_and_show(img, [xyxy])
        # Kích thước của ảnh gốc
        org_height, org_width = org_image_hw

        # Tọa độ crop trên ảnh gốc
        crop_x1, crop_y1, crop_x2, crop_y2 = crop_image
        crop_width = crop_x2 - crop_x1
        crop_height = crop_y2 - crop_y1

        # Kích thước của ảnh đã resize
        resize_width, resize_height = resize_image

        # Tọa độ bbox trong resizeImage (chuẩn hóa lại để có tọa độ thực)
        x_min_resize = xyxy[0] # * resize_width
        y_min_resize = xyxy[1] # * resize_height
        x_max_resize = xyxy[2] # * resize_width
        y_max_resize = xyxy[3] # * resize_height
        # draw_rectangles_and_show(img, [xyxy])

        # Tính tỷ lệ chuyển từ resizeImage sang cropImage
        scale_x = crop_width / resize_width
        scale_y = crop_height / resize_height

        # Tọa độ bbox trong cropImage
        x_min_crop = x_min_resize * scale_x
        y_min_crop = y_min_resize * scale_y
        x_max_crop = x_max_resize * scale_x
        y_max_crop = y_max_resize * scale_y

        # Tọa độ bbox trong orgImage
        x_min_org = x_min_crop + crop_x1
        x_max_org = x_max_crop + crop_x1

        y_min_org = y_min_crop + crop_y1
        y_max_org = y_max_crop + crop_y1

        # Chuyển tọa độ về dạng chuẩn hóa (normalized) trên orgImage
        org_x1_normalized = x_min_org / org_width
        org_y1_normalized = y_min_org / org_height
        org_x2_normalized = x_max_org / org_width
        org_y2_normalized = y_max_org / org_height

        # Output dạng tọa độ chuẩn hóa trên orgImage
        org_xyxyn = [x_min_org,y_min_org,x_max_org,y_max_org]
        #
        org_normalized=[
            org_x1_normalized ,
            org_y1_normalized ,
            org_x2_normalized ,
            org_y2_normalized 
        ]
        return org_xyxyn

    def Augment_image_in_multiple_ways(self, image, options=None):
        """
        Python use:

        ```python
        options = {
            "cover_yolo_string": "0 0.496094 0.772135 0.527344 0.453125",  # or ''
            "cover_yolo_string__resize_after_crop_WH": [800, 600],  # or false
            "Brightness": 30,  # or 0
            "Shadow_Brightness": 0,  # or 0
            "Contrast": 1.2,  # or 0
            "Saturation": 0,  # or 0
            "Hue": 0,  # or 0
            "image_path": "",
            "Change_labels": False,
        }
        AUG = ImageAugmentation()
        image, retDict = AUG.Augment_image_in_multiple_ways(image, options=options)

        retDict['newLabel']
        retDict['orgImHW']
        retDict['orgXyxy']
        ```
        """
        if options is None:
            options = {
                "cover_yolo_string": "0 0.496094 0.772135 0.527344 0.453125",  # or ''
                "cover_yolo_string__resize_after_crop_WH": [800, 600],  # or false
                "Brightness": 30,  # or 0
                "Shadow_Brightness": 0,  # or 0
                "Contrast": 1.2,  # or 0
                "Saturation": 0,  # or 0
                "Hue": 0,  # or 0
                "image_path": "",
                "Change_labels": False,
            }
        Change_labels = options.get("Change_labels", False)

        retDict={}
        cover_yolo_string=options.get("cover_yolo_string", False)
        if cover_yolo_string:
            imH, imW = image.shape[:2]
            id, x1, y1, x2, y2 = yolo_str_to_xyxy( yolo_str=mParams.cover_yolo_string, imH=imH, imW=imW )

            d = mParams.cover_yolo_string__expand_n_pixel
            y1, y2, x1, x2 = y1 - d, y2 + d, x1 - d, x2 + d
            x1 = max(x1, 0)
            y1 = max(y1, 0)
            x2 = min(x2, imW)
            y2 = min(y2, imH)
            image = image[y1:y2, x1:x2]

            if Change_labels:
                image_path = options.get("image_path", '')
                label_path = image_path.replace(".jpg", ".txt")
                if exists(label_path):
                    x_min_crop, y_min_crop = x1, y1
                    new_imH, new_imW = image.shape[:2]
                    retDict['newLabel'] = update_labels( label_path, x_min_crop, y_min_crop, new_imW, new_imH, imW, imH, )
            retDict["orgImageHW"] = (imH, imW)
            # retDict["cropImage_xyxy"] = (x1, y1, x2, y2)
            retDict["cropImage_xyxyn"] = (x1/imW, y1/imH, x2/imW, y2/imH)

            cover_yolo_string__resize_after_crop_WH = options.get("cover_yolo_string__resize_after_crop_WH", False)
            if cover_yolo_string__resize_after_crop_WH:
                image = cv2.resize( image, tuple(cover_yolo_string__resize_after_crop_WH) )

        Brightness= options.get("Brightness", 0)
        if Brightness:
            image = self.Change_Brightness(image, Brightness)

        Shadow_Brightness = options.get("Shadow_Brightness", 0)
        if Shadow_Brightness:
            image = self.Change_Shadow_Brightness(image, Shadow_Brightness)

        Contrast = options.get("Contrast", 0)
        if Contrast:
            image = self.Change_Contrast(image, Contrast)

        Saturation = options.get("Saturation", 0)
        if Saturation:
            image = self.Change_Saturation(image, Saturation)

        Hue = options.get("Hue", 0)
        if Hue:
            image = self.Change_Hue(image, Hue)

        return image, retDict

    def Change_image_multiple_ways(self, image,image_path, options):
        """
        console use
        """

        image, retDict = self.Augment_image_in_multiple_ways(image, options=options)
        if mParams.Display_image:
            taImshow(title="Augmented image", image=image, wait=None)
            keyPressed=cv2.waitKey(mParams.Display_image_Stop_to_View_in___cv2_waitkey__ms)
            if keyPressed==32:
                while True:
                    keyPressed = cv2.waitKey(
                        mParams.Display_image_Stop_to_View_in___cv2_waitkey__ms
                    )
                    if keyPressed==32:
                        break
                    if keyPressed==27:
                        os._exit(1)
        newLabel = retDict.get("newLabel", "")
        return image, newLabel

    def Augment_folder(self, **options):
        if mParams.image_folder__input == "":
            print("Hãy cấu hình giá trị cho file tham số trước, rồi chạy lại.")
            os.startfile("configs_ntanh_libs.yml")
            return
        image_folder_input = mParams.image_folder__input
        image_folder_output = mParams.image_folder_output

        FIS = mParams.fnFIS(image_folder_input, exts=(".jpg",))
        if mParams.Max_random_image_to_Aug > 0:
            FIS = [random.choice(FIS) for _ in range(mParams.Max_random_image_to_Aug)]
        nImg = len(FIS)

        if not exists(image_folder_output):
            os.makedirs(image_folder_output, exist_ok=True)

        for image_path in tqdm(FIS):
            image = cv2.imread(image_path)
            if image is None:
                print(f"Failed to read image: {image_path}")
                continue

            options ["image_path"]= image_path
            options ["Change_labels"]= True

            image_out, newLabel = self.Change_image_multiple_ways( image, image_path, options )
            fnOut = image_path.replace(image_folder_input, image_folder_output)
            label_inp = image_path.replace(".jpg", ".txt")
            label_out = fnOut.replace('.jpg', '.txt')

            os.makedirs(dirname(fnOut), exist_ok=True) 
            cv2.imwrite(fnOut, image_out)
            if mParams.Copy_label_when_save_augment_image:
                if newLabel:
                    with open(label_out, "w") as file:
                        file.write(newLabel)
                elif exists(label_inp):
                    shutil.copy(label_inp, label_out) 
            # print(".", end="", flush=True)

        print()
        print(f"Done augmenting {nImg} images to {image_folder_output}")

def Aug_Folder():
    if not mParams.Ready_to_run:
        print('Thay đổi tham số config trong file:', mParams.get_Home_Dir())
        return

    aug = ImageAugmentation()
    print("cover_yolo_string:", mParams.get("cover_yolo_string", False))
    print("Brightness:", mParams.get("Brightness", False))
    print("Shadow_Brightness:", mParams.get("Shadow_Brightness", False))
    print("Contrast:", mParams.get("Contrast", False))
    print("Saturation:", mParams.get("Saturation", False))
    print("Hue:", mParams.get("Hue", False))
    aug.Augment_folder(
        cover_yolo_string=mParams.get("cover_yolo_string", False),
        cover_yolo_string__resize_after_crop_WH=mParams.get("cover_yolo_string__resize_after_crop_WH", False),
        Brightness=mParams.get("Brightness", False),
        Shadow_Brightness=mParams.get("Shadow_Brightness", False),
        Contrast=mParams.get("Contrast", False),
        Saturation=mParams.get("Saturation", False),
        Hue=mParams.get("Hue", False),
        Vibrance=mParams.get("Vibrance", False),
    )


import cv2
import numpy as np


def draw_rectangles_and_show(
    iframe_org, bbox_list, title="Rectangles", waitkey=0, destroyWindows=True, Linewidth=2
):
    """
    Draw rectangles on image and display resized version

    Parameters:
    iframe_org: numpy array - Original input image
    bbox_list: list of [x1,y1,x2,y2] coordinates for rectangles
    """
    # Make a copy of original image to avoid modifying it
    image = iframe_org.copy()

    # Draw all rectangles on original size image
    for bbox in bbox_list:
        x1, y1, x2, y2 = [int(coord) for coord in bbox]
        cv2.rectangle(image, (x1, y1), (x2, y2), (0, 255, 0), Linewidth)

    # Resize image for display
    display_image = cv2.resize(image, (800, 600))

    # Show image
    cv2.imshow(title, display_image)
    cv2.waitKey(waitkey)
    if destroyWindows:cv2.destroyAllWindows()

    return image, display_image


def draw_xyxyn_rectangles( iframe_org, bbox_list, resizeWH=None,COLOR=(0, 255, 0), Linewidth=2):
    """
    Draw rectangles on image and display resized version

    Parameters:
    iframe_org: numpy array - Original input image
    bbox_list: list of [x1,y1,x2,y2] coordinates for rectangles
    """
    # Make a copy of original image to avoid modifying it
    image = iframe_org
    H,W=image.shape[:2]
    # Draw all rectangles on original size image
    for bbox in bbox_list:
        x1, y1, x2, y2 = bbox  
        x1, x2=int(x1*W), int(x2*W) 
        y1, y2=int(y1*H), int(y2*H)
        cv2.rectangle(image, (x1, y1), (x2, y2), COLOR, Linewidth)
    if resizeWH is not None:
        image = cv2.resize(image, resizeWH)
    return image


def show_image(image, title="Image", waitkey=1, destroyWindows=False):
    cv2.imshow(title, image)
    cv2.waitKey(waitkey)
    if destroyWindows:
        cv2.destroyAllWindows()


import cv2
import numpy as np


def draw_yolo_boxes_and_show(iframe_org, bbox_list, waitkey=0, destroyWindows=True, Linewidth=2):
    """
    Draw rectangles from YOLO format coordinates and display resized version

    Parameters:
    iframe_org: numpy array - Original input image
    bbox_list: list of [x_center, y_center, width, height] YOLO format coordinates (normalized 0-1)
    """
    # Make a copy of original image to avoid modifying it
    image = iframe_org.copy()

    # Get image dimensions
    height, width = image.shape[:2]

    # Draw all rectangles on original size image
    for bbox in bbox_list:
        # Convert YOLO format to pixel coordinates
        x_center, y_center, w, h = bbox

        # Convert normalized coordinates to pixel values
        x_center = int(x_center * width)
        y_center = int(y_center * height)
        w = int(w * width)
        h = int(h * height)

        # Calculate corner points
        x1 = int(x_center - w / 2)
        y1 = int(y_center - h / 2)
        x2 = int(x_center + w / 2)
        y2 = int(y_center + h / 2)

        # Ensure coordinates are within image bounds
        x1 = max(0, x1)
        y1 = max(0, y1)
        x2 = min(width, x2)
        y2 = min(height, y2)

        # Draw rectangle
        cv2.rectangle(image, (x1, y1), (x2, y2), (0, 255, 0), Linewidth)

    # Resize image for display
    display_image = cv2.resize(image.copy(), (800, 600))

    cv2.imshow("Rectangles", display_image)
    cv2.waitKey(waitkey)
    if destroyWindows:
        cv2.destroyAllWindows()

    return image, display_image


def convert_xyxyn_coordinates_to_new_image(xyxyn, cropImage_xyxy, HW=None):
    """Chuyển đổi hệ trục tọa độ,
    Chúng ta có:
    - ảnh Im1 là ảnh gốc, ảnh to
    - ảnh Im2 là ảnh được crop từ ảnh Im1, là 1 phần của ảnh Im1, có vùng crop là: xr1, yr1, xr2, yr2 (size: cropH, cropW)) tọa độ này thuộc tọa độ ảnh Im1
    - ảnh Im3 là ảnh được resize (nhiều lần) từ ảnh Im2 => ta có ảnh cuối cùng là Im3, mang đi predict, có box: xyxyn (tọa độ xyxy nhưng tỷ lệ 0..1)

    Hàm này chuyển xyxyn (trên ảnh Im3) về tọa độ xyxy trên ảnh Im1.

    Nếu có HW thì sẽ chuyển về tọa độ xyxyn trên ảnh Im1. Lợi ích của việc dùng xyxyn là nó không bị ảnh hưởng khi resize.

    Args:
        xyxyn: x1, y1, x2, y2 của bbox thuộc khoảng 0..1
        cropImage_xyxy: x1, y1, x2, y2 của tọa độ khi crop ảnh từ ảnh gốc.

    Returns:
        xyxy hoặc xyxyn nếu HW != None
    """
    xr1, yr1, xr2, yr2 = cropImage_xyxy # Tọa độ crop trên ảnh Im1
    cropH = yr2-yr1 
    cropW = xr2-xr1
    xn1, yn1, xn2, yn2 = xyxyn
    x1, x2 =  cropW * xn1, cropW * xn2
    y1, y2 =  cropH * yn1, cropH * yn2
    x1 = x1 + xr1
    x2 = x2 + xr1
    y1 = y1 + yr1
    y2 = y2 + yr1
    if HW is not None:
        x1, x2 = x1 / HW[1], x2 / HW[1]
        y1, y2 = y1 / HW[0], y2 / HW[0]
    return x1, y1, x2, y2


def get_Org_xyxyn__from_Cropped_resized_bboxn(xyxyn, cropImage_xyxyn):
    """Chuyển đổi hệ trục tọa độ,
    
    Quy ước: xyxyn (tọa độ xyxy nhưng tỷ lệ 0..1)
     
    Chúng ta có:
    - ảnh Im1 là ảnh gốc, ảnh to
    - ảnh Im2 là ảnh được crop từ ảnh Im1, là 1 phần của ảnh Im1, có vùng crop là: `xr1, yr1, xr2, yr2` (size: `cropH, cropW`)) tọa độ này thuộc tọa độ ảnh Im1.
      Tọa độ này được convert vể xyxyn, đặt tên là `cropImage_xyxyn` .
    - ảnh Im3 là ảnh được resize (nhiều lần) từ ảnh Im2 => ta có ảnh cuối cùng là Im3, mang đi predict, có box: xyxyn  

    Hàm này chuyển xyxyn (trên ảnh Im3) về tọa độ xyxyn trên ảnh Im1. Lợi ích của việc dùng xyxyn là nó không bị ảnh hưởng khi resize.

    Args:
        xyxyn: x1, y1, x2, y2 của bbox (thuộc khoảng 0..1)
        cropImage_xyxyn: x1, y1, x2, y2 của tọa độ khi crop ảnh từ ảnh gốc (thuộc khoảng 0..1).

    Returns:
        xyxyn 
    """
    xr1, yr1, xr2, yr2 = cropImage_xyxyn  # Tọa độ crop trên ảnh Im1
    x1, y1, x2, y2 = xyxyn
    x1 = x1 * (xr2 - xr1) + xr1
    x2 = x2 * (xr2 - xr1) + xr1
    y1 = y1 * (yr2 - yr1) + yr1
    y2 = y2 * (yr2 - yr1) + yr1
    xyxyn = (x1, y1, x2, y2)
    return xyxyn


if __name__ == "__main__":
    RUN=2
    if RUN==1:
        Aug_Folder()
    if RUN==2:
        import numpy as np
        aug = ImageAugmentation()
        iframe_org = cv2.imread("d:/iframe_org.jpg")
        with open("d:/retDict.pkl", "rb") as handle:
            retDict = pickle.load(handle)

        with open("d:/results.pkl", "rb") as handle:
            results = pickle.load(handle)

        # =========================================
        for k, box in enumerate(results.boxes):
            # x1, y1, x2, y2 = np.array(box.xyxy.cpu(), dtype=np.dtype(np.int_)).squeeze()
            xyxyn = list( tuple(np.array(box.xyxyn.cpu(), dtype=np.dtype(np.float_)).squeeze()) )
            xyxy = list( tuple(np.array(box.xyxy.cpu(), dtype=np.dtype(np.float_)).squeeze()) )

            confidence = float(box.conf.cpu())

            img = results[k].plot()  # This plots the detections on the image
            print("{}".format(img.shape,))
            # Convert BGR to RGB (OpenCV uses BGR by default)
            # img_rgb = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
            x1, y1, x2, y2=xyxy
            draw_rectangles_and_show(img, [[x1, y1, x2, y2]], title='aaa111', waitkey=1, destroyWindows=False, Linewidth=2)
            # cv2.imshow('Detection Results', img)
            # cv2.waitKey(1)

            x1,y1,x2,y2 = retDict["cropImage_xyxy"]
            H,W=retDict["orgImageHW"]
            x1, x2 = x1/W, x2/W
            y1, y2 = y1/H, y2/H
            retDict["cropImage_xyxyn"] =(x1,y1,x2,y2)
            restore_options = {
                "orgImageHW": retDict["orgImageHW"],
                # "cropImage_xyxy": retDict["cropImage_xyxy"], 
                "cropImage_xyxyn": retDict["cropImage_xyxyn"], 
                "resizeImageHW": retDict["resizeImage"],
                "xyxyn": xyxyn,
                # "xyxy": xyxy,
            }
            print(restore_options)
            # ret = aug.restore_coordinates(iframe_org, restore_options)
            H_predict, W_predict = results.orig_shape
            H, W = restore_options["orgImageHW"]
            # x_center, y_center, box_width, box_height = xyxyn
            # x1, y1, x2, y2 = retDict["cropImage_xyxy"]
            # draw_rectangles_and_show(iframe_org, [[x1, y1, x2, y2]], title='aaa', waitkey=1, destroyWindows=False, Linewidth=4)
            # draw_rectangles_and_show(
            #     img, [xyxy], title="bbbb", waitkey=1, destroyWindows=False
            # )

            # ret = convert_xyxyn_coordinates_to_new_image( xyxyn=xyxyn, cropImage_xyxy=retDict["cropImage_xyxy"], HW=restore_options["orgImageHW"] )

            # print(ret)
            # print()
            # bbox_list = [ret]
            # # draw_rectangles_and_show(iframe_org, bbox_list)
            # draw_rectangles_and_show(iframe_org, bbox_list, Linewidth=5,waitkey=1,)
            cropImage_xyxyn = retDict["cropImage_xyxyn"]
            ret1 = get_Org_xyxyn__from_Cropped_resized_bboxn(xyxyn, cropImage_xyxyn)
            pp(ret1)
            draw_xyxyn_rectangles_and_show(iframe_org, [ret1], COLOR=(0,0,255), Linewidth=3)

        # print(ret)
# cv2.imshow("img", iframe[70:106,0:5]);cv2.waitKey(0)



import textwrap
import cv2
from PIL import ImageFont, Image, ImageDraw
import numpy as np

def fnPlot_Warp_Text(img, lstlines, x1, y1, twidth=35, font_size=0.7, font_thickness=1, lineSpacing=10, color=(0, 0, 0), bg_Width=4):
    font = cv2.FONT_HERSHEY_SIMPLEX
    if type(lstlines) == str:
        # Tách các dòng trong xâu text bằng ký tự xuống dòng (\n)
        lstlines = lstlines.splitlines()
    
    # Khởi tạo tọa độ y ban đầu cho dòng đầu tiên
    y = y1
    textsize = None
    # Lặp qua từng dòng và in trên ảnh
    for i, line in enumerate(lstlines):
        wrapped_text = textwrap.wrap(line, width=twidth)
        if len(line) == 0:
            wrapped_text = ['']
        for j, wrapped_line in enumerate(wrapped_text):
            if textsize is None:
                textsize = cv2.getTextSize(wrapped_line, font, font_size, font_thickness)[0]
            # gap = textsize[1] + lineSpacing
            
            # Cập nhật tọa độ y cho dòng mới
            y += textsize[1] + lineSpacing
            
            x = int(x1)
            cv2.putText(img, wrapped_line, (x, int(y)), font, font_size,  # (128, 128, 128),
                        (255, 255, 255),
                        bg_Width,
                        lineType=cv2.LINE_AA
                        )
            cv2.putText(img, wrapped_line, (int(x), int(y)), font, font_size, color, font_thickness, lineType=cv2.LINE_AA)
    return img


def PlotText_UTF8(image, text, x=50, y=50, font="C:/Windows/Fonts/arial.ttf", font_size=35, color=(255, 0, 0, 255), change_color_position=False):
    """
    Plot text using UTF-8 (vietnamese)
    """
    if len(color) == 3:
        color = tuple(list(color) + [255])
    if change_color_position:
        b = color
        b = (b[2], b[1], b[0], b[3])
        color = b
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    pil_image = Image.fromarray(image)
    # Draw non-ascii text onto image
    font = ImageFont.truetype(font, font_size)
    draw = ImageDraw.Draw(pil_image)
    draw.text((x, y), text, font=font, fill=color)
    # Convert back to Numpy array and switch back from RGB to BGR
    image = np.asarray(pil_image)
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    return image


def PlotText_UTF8_outline(image, text, x=50, y=50, font="C:/Windows/Fonts/arial.ttf", font_size=30, color=(255, 0, 0, 255), outline_color=(255, 255, 255, 128), outline_width=1):
    """
    Plot text using UTF-8 (vietnamese) with an outline
    """
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    pil_image = Image.fromarray(image)
    
    # Draw non-ascii text onto image with outline
    font = ImageFont.truetype(font, font_size)
    draw = ImageDraw.Draw(pil_image)
    
    # Draw the outline
    for dx in range(-outline_width, outline_width + 1):
        for dy in range(-outline_width, outline_width + 1):
            draw.text((x + dx, y + dy), text, font=font, fill=outline_color)
    
    # Draw the main text
    draw.text((x, y), text, font=font, fill=color)
    
    # Convert back to Numpy array and switch back from RGB to BGR
    image = np.asarray(pil_image)
    image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    return image


def opposite_color(color):
    return tuple(255 - value for value in color)


def putTextWithOutline(image, text, position, font, font_scale, color, thickness):
    # outline_color = (128, 128, 128)
    outline_color = opposite_color(color)
    img = cv2.putText( image, text, position, font, font_scale, outline_color, thickness + 3, cv2.LINE_AA, )
    img = cv2.putText( img, text, position, font, font_scale, color, thickness, cv2.LINE_AA )
    return img


def draw_bounding_box( image, list_x1y1x2y2, label_text="", font_scale=None, thickness=None, TopOnly=False, colour=(0, 255, 0), hesoFont=1.0, hesoLine=1.0, ):
    # Calculate thickness and font scale based on image width
    h, w = image.shape[:2]
    if not thickness:
        thickness = int(max(1, hesoLine * h / 420))
    if not font_scale:
        font_scale = max(0.5, hesoFont * h / 1200.0)
    # Draw the bounding box
    x1, y1, x2, y2 = list_x1y1x2y2
    if TopOnly:
        y2 = y1 + thickness
    cv2.rectangle(image, (x1, y1), (x2, y2), (188, 188, 188), thickness + 1)
    cv2.rectangle(image, (x1, y1), (x2, y2), colour, thickness)
    # Put the label and confidence

    cv2.putText( image, label_text, (x1, y1 - 10), cv2.FONT_HERSHEY_DUPLEX, font_scale, (120, 120, 120), thickness + 1, )
    cv2.putText( image, label_text, (x1, y1 - 10), cv2.FONT_HERSHEY_DUPLEX, font_scale, colour, thickness, )
    return image


def PlotText(image, label_text, x1, y1, font_scale, thickness, colour=(0, 255, 0)):
    cv2.putText( image, label_text, (x1, y1), cv2.FONT_HERSHEY_DUPLEX, font_scale, (120, 120, 120), thickness + 1, )
    cv2.putText( image, label_text, (x1, y1), cv2.FONT_HERSHEY_DUPLEX, font_scale, colour, thickness, )



APP_NAME = "image_dupplicate_remove"
import os, sys
import os.path
from pprint import pprint as pp
from os.path import dirname, basename, join, isfile, isdir,exists
import time
# import imagehash
import numpy as np
from tqdm import tqdm
# print(os.getcwd()) # Thư mục console đang chạy
# print(__file__) # Thư mục file code
codeDir = dirname(os.path.abspath(__file__))
sys.path.append(codeDir)
# pp(sys.path)
# from PIL import Image, ImageChops
from ParamsBase import tactParametters as BasePr
from image_augmentation import taImshow
import shutil


class Parameters1(BasePr):
    def __init__(self, ModuleName="TACT"):
        super().__init__(saveParam_onlyThis_APP_NAME=True)
        # self.onlyThisAPP_NAME=True
        self.Ready_to_run = False
        self.AppName = APP_NAME
        self.Intro = "Chương trình này dành riêng cho việc remove ảnh gần giống nhau, phục vụ cho mục đích training model"
        self.HD_CMD = "ntanh_img_del"
        self.HD_install_this_lib = "pip install ntanh"
        self.HD = {
            "N_first_files_image_to_find_duplicate=x": "x==0: chạy hết tất cả files, x>0: chỉ chạy x file đầu tiên",
            "dupplicate_threshold": "Số thực, cần căn cứ vào số sai khác thực tế theo Diff histogram in cuối mỗi lần chạy",
            "num_diff_average_bins": "Số lượng bị cần chia khi tính histogram, bin càng nhiều thì chia Diff càng nhỏ",
            "Run_actual_move_file": "=True: nó sẽ thực sự chuyển các file vào image_folder_output khi diff<dupplicate_threshold, =false: chỉ tính, không chuyển file, gọi là dry run.",
        }
        self.dupplicate_threshold = 5
        self.image_folder__input = ""
        self.image_folder_output = ""
        self.Loop_Move_image_and_label_until_no_duplicate = False
        self.N_first_files_image_to_find_duplicate = 10
        self.num_diff_average_bins=10
        self.Run_actual_move_file=True
        self.load_then_save_to_yaml(file_path=f"{APP_NAME}.yml", ModuleName=ModuleName)


mParams = Parameters1(APP_NAME)
import os
import shutil
from PIL import Image
from skimage.metrics import structural_similarity as ssim
from skimage.io import imread
from skimage.transform import resize


class ImageDuplicateRemover:
    def __init__(self):
        pass

    def fnRemove(self):
        # Ngưỡng SSIM để xác định ảnh trùng lặp
        duplicate_threshold = mParams.dupplicate_threshold

        # Đường dẫn đến thư mục input và output
        inpPath = mParams.image_folder__input
        outPath = mParams.image_folder_output

        # Hàm lấy danh sách các file image trong thư mục input
        fis = mParams.fnFIS(inpPath)
        if mParams.N_first_files_image_to_find_duplicate>0:
            fis = fis[: mParams.N_first_files_image_to_find_duplicate]
        # Danh sách các ảnh trùng lặp
        list_duplicate_imgs = []
        checked_imgs = []

        def is_Similar__mse(imageA, imageB, threshold, name1='', name2=''):
            # the 'Mean Squared Error' between the two images is the
            # sum of the squared difference between the two images;
            # NOTE: the two images must have the same dimension
            err = np.sum((imageA.astype("float") - imageB.astype("float")) ** 2)
            err /= float(imageA.shape[0] * imageA.shape[1])

            # return the MSE, the lower the error, the more "similar"
            # the two images are
            # print(err,"\t", name1 + "===" + name2)
            return  err < threshold, err
        average = 0
        n = 0
        average_data = []
        # Duyệt qua tất cả các ảnh để so sánh
        with tqdm(total=len(fis)//2) as pbar:
            # for i in tqdm(range(0,len(fis)-1, 2)):
            for i in range(0,len(fis)-1, 2):
                # print()
                img1_path = fis[i]
                # img1 = Image.open(img1_path)
                img1 = imread(img1_path)
                name1=os.path.basename(img1_path)
                # Xử lý ảnh nếu kích thước quá nhỏ
                # img1 = preprocess_image(img1)

                j=i+1
                img2_path = fis[j]
                # img2 = Image.open(img2_path)
                img2 = imread(img2_path)
                name2 = os.path.basename(img2_path)
                # Xử lý ảnh nếu kích thước quá nhỏ
                # img2 = preprocess_image(img2)
                is_Similar, mse= is_Similar__mse(img1, img2, duplicate_threshold, name1, name2)
                average = (n * average + mse) / (n + 1)
                n += 1
                if mse < average:
                    average_data.append(mse)

                if is_Similar:
                    list_duplicate_imgs.append(img2_path)
                # So sánh hai ảnh bằng SSIM
                # if images_are_similar(img1, img2, duplicate_threshold):
                #     list_duplicate_imgs.append(img2_path)
                # print(f"Found duplicate: {img2_path} (similar to {img1_path})")
                pbar.set_postfix(
                    duplicate=f"{len(list_duplicate_imgs)} files, average diff thresh: {average:3.6f}/{duplicate_threshold}"
                )
                # Phần code xử lý
                pbar.update(1)
        NfilesRemove = len(list_duplicate_imgs)

        # Di chuyển các ảnh trùng lặp vào thư mục output
        for k,img_path in enumerate(list_duplicate_imgs):
            dst=img_path.replace(inpPath, outPath)
            if mParams.Run_actual_move_file:
                dst_Dir=dirname(dst)
                os.makedirs(dst_Dir, exist_ok=True)
                shutil.move(img_path, dst_Dir)
                label=img_path.replace(".jpg", ".txt")
                if exists(label):
                    shutil.move(label, dst_Dir)
                print(f"{k:>3}. Moved duplicate image: {img_path} \t ==> \t {outPath}")
            else:
                print(f"{k:>3}. Will move duplicate image: {img_path} \t ==> \t {outPath}")
        min_value = min(average_data)
        max_value = max(average_data)
        bins = np.linspace(min_value, max_value, mParams.num_diff_average_bins + 1)
        # Tính histogram tích lũy
        hist, bin_edges = np.histogram(average_data, bins=bins)
        # Tạo chuỗi kết quả dưới dạng (bin_value: count)
        result_str = ",".join([f"{bin_edges[j]:.2f}-{bin_edges[j+1]:.2f}: {hist[j]}" for j in range(len(hist))])
        # Hiển thị chuỗi kết quả
        result_str = result_str.split(",")
        print()
        print(f"Diff histogram [mse<average], his min-his max: Number of files:")
        pp(result_str)
        print("Căn cứ vào his này, để chọn lại tham số dupplicate_threshold trong file config cho phù hợp.")
        return NfilesRemove

def fnImage_dupplicate_remove():
    if not mParams.Ready_to_run:
        print("Thay đổi tham số config trong file:", mParams.get_Home_Dir())
        return
    start = time.time()
    # Sử dụng lớp ImageDuplicateRemover
    image_remover = ImageDuplicateRemover()
    NfilesRemove=image_remover.fnRemove()
    if mParams.Loop_Move_image_and_label_until_no_duplicate:
        while NfilesRemove>0:
            NfilesRemove = image_remover.fnRemove()
    print("Total running:", time.time() - start, "s")

if __name__ == "__main__":
    fnImage_dupplicate_remove()



import cv2
import numpy as np 

def rotate_image(image, angle):
    image_center = tuple(np.array(image.shape[1::-1]) / 2)
    rot_mat = cv2.getRotationMatrix2D(image_center, angle, 1.0)
    result = cv2.warpAffine(image, rot_mat, image.shape[1::-1], flags=cv2.INTER_LINEAR)
    return result
 

def ImageInfomation(img):
    # Calculate the focus measure using Laplacian operator
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    focus_measure = cv2.Laplacian(gray, cv2.CV_64F).var()

    # Calculate the brightness using mean of pixel values
    brightness = cv2.mean(img)[0]

    # Calculate the contrast using standard deviation of pixel values
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    contrast = cv2.meanStdDev(gray)[1][0][0]

    # Calculate the sharpness using Laplacian of Gaussian operator
    gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    blur = cv2.GaussianBlur(gray, (5, 5), 0)
    sharpness = cv2.Laplacian(blur, cv2.CV_64F).var()

    # Print the calculated parameters
    # print('Focus:', focus_measure)
    # print('Brightness:', brightness)
    # print('Contrast:', contrast)
    # print('Sharpness:', sharpness)
    return f"Focus:{round(focus_measure, 1)}, brightness:{round(brightness)}, contrast:{round(contrast)}, sharpness:{round(sharpness)}"




import cv2
import numpy as np

from ntanh.yolo_boxes import YoloCrop


def calculate_luminance(image, Lumen_coefficient=1.0):
    """
    Hàm tính độ sáng trung bình từ một ảnh đầu vào.

    :param image_path: Đường dẫn tới ảnh.
    :return: Độ sáng trung bình.
    """

    # Chuyển đổi ảnh sang không gian màu YUV
    yuv_image = cv2.cvtColor(image, cv2.COLOR_BGR2YUV)

    # Kênh Y chứa độ sáng
    luminance_channel = yuv_image[:, :, 0]

    # Tính độ sáng trung bình
    average_luminance = np.mean(luminance_channel)

    return average_luminance * Lumen_coefficient


if __name__ == "__main__":
    RUN=1
    # Lumen calculation: ---------------------------------------------------------------------
    if RUN==1:
        coverYolo_str = "0 0.496094 0.772135 0.527344 0.453125"

        # Ví dụ
        image_path = (
            r"H:\DATA\Cam360SmartGate\Training_data_from_DUY\_Train\11_10\6\6_1\frame000000.jpg"
        )
        # Đọc ảnh
        image = cv2.imread(image_path)

        image = YoloCrop(image, coverYolo_str, Delta_Expand=0)
        luminance = calculate_luminance(image)        
        print(f"{luminance:0.4f} (luminance)")

        cv2.imshow("Before", image)
        # Ví dụ
        image_path = r"F:\Training_data_from_DUY\_Train\11_10\6\6_1\frame000000.jpg"
        # Đọc ảnh
        image = cv2.imread(image_path)
        image = YoloCrop(image, coverYolo_str, Delta_Expand=0)
        luminance = calculate_luminance(image)
        print(f"After {luminance:0.4f} (lumen)")
        cv2.imshow("after", image)
        cv2.waitKey(0)
    # End Lumen calculation: ---------------------------------------------------------------------
    if RUN==2:
        pass



def compare_labels(label_list, rule_dict):
    """So sánh list của các nhãn đã predict với dict chuẩn được khái báo trước.
        Lấy label yHat so với y, để đưa ra kết luận OK/NG. Logic như sau:

        -------------------------------------------------------------------------------

        class mParams:
            detect_abnormal_Results_Conclusion = {
                "4": {"Min_num": 0, "Max_num": 1, "name": "Blue"},
                "5": {"Min_num": 0, "Max_num": 1, "name": "White"},
                "6": {"Min_num": 0, "Max_num": 1, "name": "Black"},
                "7": {"Min_num": 0, "Max_num": 0, "name": "No coach"},
                "8": {"Min_num": 0, "Max_num": 0, "name": "Bend"},
                "Sum_same_labels_min_max": {
                    "group1": {"labels": "4,5,6", "min": 0, "max": 1},
                    "group2": {"labels": "3", "min": 0, "max": 1},
                },
            }
            hide_draw_labels = ["3"]

        # Danh sách cần kiểm tra
        list_Need_to_check = [
            [4],  # OK
            [5],  # OK
            [6],  # OK
            [4, 5],  # OK
            [4, 5, 6],  # OK
            [7],  # NG
            [8],  # NG
            [4, 4],  # NG
            [4, 4, 4],  # NG
            [4, 5],  # NG
            [5, 5],  # NG
            [5, 5, 4],  # NG
            [5, 5, 4, 4],  # NG
            [4, 5, 6],  # NG
            [4, 5, 6, 7],  # NG
            [4, 8],  # NG
            [4, 7],  # NG
        ]

        # Thực hiện kiểm tra và in kết quả
        for label1 in list_Need_to_check:
            kq = compare_labels(label1, mParams.detect_abnormal_Results_Conclusion)
            print(label1, kq)

        return:
            f"NG, {'; '.join(violations)}"
            f"OK, {rule_dict[label]['name']}"
        -------------------------------------------------------------------------------

        So sánh nhãn trong `label_list` với quy tắc `Min_num` và `Max_num` từ `rule_dict`.

        Args:
            label_list (list): Danh sách các nhãn cần kiểm tra.
            rule_dict (dict): Từ điển chứa quy tắc kiểm tra các nhãn với:
                - `Min_num`: Số lượng tối thiểu mà nhãn cần phải xuất hiện.
                - `Max_num`: Số lượng tối đa mà nhãn được phép xuất hiện.
                - `name`: Tên của nhãn để hiển thị trong kết quả.

        Returns:
            str: Kết quả kiểm tra theo các trường hợp sau:
                - "OK, name" nếu tất cả các nhãn trong `label_list` thỏa mãn điều kiện Min_num và Max_num.
                - "NG, lý do" nếu có bất kỳ nhãn nào vi phạm các điều kiện:
                    - Số lượng nhãn vượt quá `Max_num`.
                    - Số lượng nhãn ít hơn `Min_num`.
                    - Nhãn yêu cầu xuất hiện ít nhất một lần (Min_num > 0) nhưng không có trong `label_list`.

        Logic:
        - Đầu tiên, đếm số lần xuất hiện của mỗi nhãn trong `label_list`.
        - Đối với mỗi nhãn trong `label_list`, kiểm tra xem số lần xuất hiện có thỏa mãn:
            - Không vượt quá `Max_num`.
            - Không ít hơn `Min_num`.
        - Nếu có nhãn nào yêu cầu số lượng tối thiểu (`Min_num > 0`) nhưng không xuất hiện trong `label_list`,
          thì ghi nhận vi phạm đó.
        - Nếu có bất kỳ vi phạm nào (vượt quá `Max_num`, ít hơn `Min_num` hoặc nhãn bị thiếu), trả về "NG" kèm theo lý do.
        - Nếu tất cả các nhãn đều thỏa mãn, trả về "OK" và tên nhãn đầu tiên trong `label_list`.
    """
    # Đếm số lượng mỗi nhãn xuất hiện trong label_list
    label_count = {}
    for label in label_list:
        str_label = str( label )  # Chuyển label sang chuỗi để khớp với key trong rule_dict
        if str_label in label_count:
            label_count[str_label] += 1
        else:
            label_count[str_label] = 1

    # Danh sách lưu các nhãn vi phạm
    violations = []

    # Kiểm tra từng nhãn dựa trên quy tắc trong rule_dict
    for label, count in label_count.items():
        if label in rule_dict:
            min_num = rule_dict[label].get("Min_num", 0)
            max_num = rule_dict[label]["Max_num"]
            name = rule_dict[label]["name"]

            # Kiểm tra vi phạm Max_num
            if count > max_num:
                if max_num > 0:
                    violations.append(f"{name} > {max_num}")
                else:
                    violations.append(f"{name}")

            # Kiểm tra vi phạm Min_num
            if count < min_num:
                if min_num > 0:
                    violations.append(f"{name} < {min_num}")
                else:
                    violations.append(f"{name}")

    # Kiểm tra các nhãn còn thiếu (Min_num yêu cầu nhưng không có trong danh sách)
    for label, rules in rule_dict.items():
        if label == "Sum_same_labels_min_max":
            continue
        if str(label) not in label_count and rules.get("Min_num", 0) > 0:
            violations.append(f"{rules['name']} < {rules['Min_num']}")

    if "Sum_same_labels_min_max" in rule_dict:
        for grp_key, vl_key in rule_dict["Sum_same_labels_min_max"].items():
            # grp_key="group1"
            # vl_key = {"labels": "7,8", "min": 0, "max": 0}
            gLabels = vl_key["labels"].split(",")
            gmin = vl_key["min"]
            gmax = vl_key["max"]
            predLabels_cnt = 0
            for label, count in label_count.items():
                if label in gLabels:
                    predLabels_cnt += count
            if predLabels_cnt < gmin or predLabels_cnt > gmax:
                violations.append(
                    f"Grp: {vl_key['labels']}: {predLabels_cnt} not in [{gmin},{gmax}]"
                )

    # Nếu có vi phạm, trả về NG và liệt kê lý do
    if violations:
        return f"NG, {'; '.join(violations)}"

    # Nếu không có nhãn nào vi phạm
    # Lấy tên của nhãn đầu tiên trong danh sách label_list
    lbName = "OK, -"
    for label in label_list:
        if label in rule_dict:
            if "name" in rule_dict[label]:
                lbName = f"OK, {rule_dict[label]['name']}"
                break
    return lbName


if __name__ == "__main__":

    class mParams:
        detect_abnormal_Results_Conclusion = {
            "4": {"Min_num": 0, "Max_num": 1, "name": "Blue"},
            "5": {"Min_num": 0, "Max_num": 1, "name": "White"},
            "6": {"Min_num": 0, "Max_num": 1, "name": "Black"},
            "7": {"Min_num": 0, "Max_num": 0, "name": "No coach"},
            "8": {"Min_num": 0, "Max_num": 0, "name": "Bend"},
            "Sum_same_labels_min_max": {
                "group1": {"labels": "4,5,6", "min": 0, "max": 1},
                "group2": {"labels": "3", "min": 0, "max": 1},
            },
        }
        hide_draw_labels = ["3"]

    # Danh sách cần kiểm tra
    list_Need_to_check = [
        [4],  # OK
        [5],  # OK
        [6],  # OK
        [4, 5],  # OK
        [4, 5, 6],  # OK
        [7],  # NG
        [8],  # NG
        [4, 4],  # NG
        [4, 4, 4],  # NG
        [4, 5],  # NG
        [5, 5],  # NG
        [5, 5, 4],  # NG
        [5, 5, 4, 4],  # NG
        [4, 5, 6],  # NG
        [4, 5, 6, 7],  # NG
        [4, 8],  # NG
        [4, 7],  # NG
    ]

    # Thực hiện kiểm tra và in kết quả
    for label1 in list_Need_to_check:
        kq = compare_labels(label1, mParams.detect_abnormal_Results_Conclusion)
        print(label1, kq)



from collections import deque
def test():
    print("Test, Labels_Filter at:", __file__)

# ==========================================================================================

class LabelFilter_1day_so:
    def __init__(self, min_consecutive):
        self.min_consecutive = min_consecutive
        self.buffer = deque(maxlen=min_consecutive)
        self.valid_values = deque(maxlen=min_consecutive)
        self.result = deque(maxlen=min_consecutive)

    def process_value(self, value):
        self.buffer.append(value)
        if len(self.buffer) == self.min_consecutive and len(set(self.buffer)) == 1:
            self.valid_values.append(value)
        if value in set(self.valid_values):
            self.result.append(value)
        if len(self.result) > 0:
            return self.result[-1]
        return []

# ==========================================================================================

class clsLabelFilter_list_bboxes:
    def __init__(self, min_consecutive):
        self.min_consecutive = min_consecutive
        self.buffer = deque(maxlen=min_consecutive)  # Lưu trữ các frame gần nhất
        self.valid_values = deque( maxlen=min_consecutive )  # Lưu trữ kết quả hợp lệ gần nhất
        self.result = None  # Kết quả ổn định gần nhất

    def process_value(self, value, debug=False):
        label_ids = tuple( sorted([item[0] for item in value]) )  # Lấy danh sách các label_id và sắp xếp
        self.buffer.append(label_ids)  # Thêm danh sách label_id vào buffer
        if debug:
            for ids in self.buffer:
                print(self.buffer[0], '==', ids )
        # Kiểm tra xem có đủ frame trong buffer và tất cả frame đều có cùng số lượng label_id hay không
        if len(self.buffer) == self.min_consecutive and all( self.buffer[0] == ids for ids in self.buffer ):
            self.valid_values.append(value)  # Nếu đúng thì cập nhật valid_values
            self.result = value  # Cập nhật kết quả ổn định gần nhất

        # Nếu có kết quả hợp lệ, trả về kết quả ổn định gần nhất
        if self.result:
            return self.result
        return []  # Nếu không có kết quả hợp lệ, trả về rỗng

# ==========================================================================================

class clsFilter_Frames_sep_by_labels:
    """
    Filter bbox trong frames: [(2, 4, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (3, 0.52, 0.51, 0.2, 0.2)]
    Chương trình sẽ tách các labels ra, và fileter theo từng nhóm labels đó, cùng labels vào 1 nhóm, có windows riêng cho mỗi label
    Cách dùng:
        class Param:
            def __init__(self, ):
                self.ClassNames = {
                    0: "Blue Hat",
                    1: "W Hat",
                    2: "Black Hat",
                    3: "Person",
                    4: "Blue",
                    5: "White",
                    6: "Black",
                    7: "No Shirt",
                    8: "Bend",
                }
                self.remove_outliers_online_windows_len_by_class_name = {
                    "default": 3,
                    0: 2,
                    1: 2,
                    2: 4,
                    3: 2,
                    4: 2,
                    5: 2,
                    6: 2,
                    7: 2,
                    8: 1,
                }
        mParams=Param()
        mFilter_Frames = Filter_Frames_sep_by_labels(ClassNames=mParams.ClassNames, windows_len_by_class=mParams.remove_outliers_online_windows_len_by_class_name)
        for cnt,frame1 in enumerate(frame): # Lặp qua các lần predict khác nhau.
            print(cnt, ">", frame1)
            vl=mFilter_Frames.Update(frame1, stt=cnt)
            print(cnt,'-', vl)
            print()
    """
    def __init__(self, ClassNames, windows_len_by_class):
        self.FilLabel={}
        self.ThisTime_bbox = {}
        self.ClassName = ClassNames
        for key, val in ClassNames.items(): 
            if key in windows_len_by_class.keys():
                thisWindow = windows_len_by_class[key]
            else:
                thisWindow = windows_len_by_class["default"]

            self.FilLabel[key]=clsLabelFilter_list_bboxes(min_consecutive=thisWindow)
            self.ThisTime_bbox [key]=[]
        self.Saved_thistime=self.ThisTime_bbox.copy()

    def ThisTime_bbox_clear(self):
        self.ThisTime_bbox=self.Saved_thistime.copy()

    def Update(self, frameBoxes, stt=None, debug=False):
        if debug:
            print()            
            print(stt, ":", frameBoxes)

        self.ThisTime_bbox_clear()
        labels = set([bbox[0] for bbox in frameBoxes]) #[(2, 0, 0.1, 0.2, 0.2), (2, 0.5, 0.5, 0.2, 0.2)]
        for label1 in labels:
            mBBox = []
            for bbox in frameBoxes:
                label_id, x, y, w, h = bbox
                if label_id == label1:
                    mBBox.append(bbox)
            self.ThisTime_bbox[label1]= mBBox

        # Filter this frame
        ret=[]
        for label_id, frame in self.ThisTime_bbox.items():
            # if frame:
            #     print(label_id, ":" * label_id, frame)
            vl = self.FilLabel[label_id].process_value(frame)
            ret.extend(vl)
            if debug: print(label_id, ":" * label_id, vl)
        return ret


# ==========================================================================================
def tim_so_dong_nhat(my_set):
    # Chuyển set thành list để có thể đếm số lần xuất hiện
    my_list = list(my_set)

    # Sử dụng từ điển để đếm tần suất
    so_lan_xuat_hien = {}
    for x in my_list:
        if x in so_lan_xuat_hien:
            so_lan_xuat_hien[x] += 1
        else:
            so_lan_xuat_hien[x] = 1

    # Tìm phần tử có số lần xuất hiện nhiều nhất
    gia_tri_lon_nhat = max(so_lan_xuat_hien.values())
    for k, v in so_lan_xuat_hien.items():
        if v == gia_tri_lon_nhat:
            return k

# # Ví dụ sử dụng:
# my_set = {1, 2, 3, 2, 4, 2, 5}
# ket_qua = tim_so_dong_nhat(my_set)
# print(f"Số xuất hiện nhiều nhất là: {ket_qua}")
# ==========================================================================================

def getframe1_keys(bboxes):
    """
    bboxes=[(22, 0, 0.1, 0.2, 0.2), (33, 0.5, 0.5, 0.2, 0.2)]
    ret   ='22,33'
    """
    thisF=[]
    for x in bboxes: 
        thisF.append(str(x[0])) 
    thisF.sort()
    ret=','.join(thisF)
    return ret


def GetSet_From_deque(mDeque):
    """
    mDeque=[(22, 0, 0.1, 0.2, 0.2), (33, 0.5, 0.5, 0.2, 0.2)], [(44, 1, 0.12, 0.21, 0.21), (55, 0.5, 0.5, 0.2, 0.2)]
    ret={'22,33','44,55'}
    """
    mDeque_values_keys=[]
    for frame1 in list(mDeque):
        thisF=getframe1_keys(bboxes=frame1)
        mDeque_values_keys.append(thisF)
    return set(mDeque_values_keys)

def fnRemove_outliers_online(min_consecutive=3):    
    buffer = deque(maxlen=min_consecutive*2)
    valid_values = deque(maxlen=min_consecutive)
    result=['']

    def process_value(value):        
        N=0
        if len(buffer) > 0:
            N = len([x[0] for x in buffer[-1]])
        buffer.append(value)        
        buffer_values_keys = GetSet_From_deque(buffer)
        # print('buffer_values_keys:', buffer_values_keys)
        if len(buffer) == min_consecutive and len(buffer_values_keys) == 1:
            valid_values.append(value)

        valid_values_keys = GetSet_From_deque( valid_values)
        # print("valid_values_keys:", valid_values_keys)

        thisf=getframe1_keys(value)
        if thisf in valid_values_keys:
            result[0]=value
        return result[0]

    return process_value


if __name__ == "__main__":

    # Sample frames with (label_id, x, y, w, h)
    if 1:
        frame=[]
        frame.append( [(2, 0, 0.1, 0.2, 0.2), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 1, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 2, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 3, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 4, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2)])
        frame.append( [(3, 5, 0.1, 0.2, 0.2),(3, 5, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2)])
        frame.append( [(2, 6, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2)])
        frame.append( [(2, 7, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2)])
        frame.append( [(2, 8, 0.1, 0.2, 0.2)] )
        frame.append( [(2, 9, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.52, 0.52, 0.2, 0.2)])
        frame.append( [(2, 10, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.54, 0.51, 0.2, 0.2)])
        frame.append( [(2, 11, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.54, 0.51, 0.2, 0.2), (2, 0.54, 0.51, 0.2, 0.2)])
        frame.append( [(2, 12, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.56, 0.53, 0.2, 0.2)])
        frame.append( [(2, 13, 0.1, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2), (2, 0.52, 0.51, 0.2, 0.2)])
        frame.append([(4, 14, 0.12, 0.21, 0.21), (2, 14, 0.12, 0.21, 0.21), (4, 14, 0.12, 0.21, 0.21), (5, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(4, 14, 0.12, 0.21, 0.21), (2, 14, 0.12, 0.21, 0.21), (4, 14, 0.12, 0.21, 0.21), (5, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(4, 14, 0.12, 0.21, 0.21), (2, 14, 0.12, 0.21, 0.21), (4, 14, 0.12, 0.21, 0.21), (5, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(4, 14, 0.12, 0.21, 0.21), (2, 14, 0.12, 0.21, 0.21), (4, 14, 0.12, 0.21, 0.21), (5, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(4, 14, 0.12, 0.21, 0.21), (2, 14, 0.12, 0.21, 0.21), (4, 14, 0.12, 0.21, 0.21), (5, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(4, 14, 0.12, 0.21, 0.21), (2, 14, 0.12, 0.21, 0.21), (4, 14, 0.12, 0.21, 0.21), (5, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 15, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 16, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [])
        frame.append( [(2, 17, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 18, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 19, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 20, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append( [(2, 21, 0.51, 0.2, 0.2)])
        frame.append( [(2, 22, 0.52, 0.21, 0.21)])
        frame.append( [(2, 23, 0.52, 0.21, 0.21)])
        frame.append( [(2, 24, 0.52, 0.21, 0.21)])
        frame.append([(2, 17, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(2, 18, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(2, 19, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(2, 19, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(2, 190, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(2, 20, 0.12, 0.21, 0.21), (2, 0.5, 0.5, 0.2, 0.2)])
        frame.append([(2, 21, 0.51, 0.2, 0.2)])
        frame.append([(2, 22, 0.52, 0.21, 0.21)])
        frame.append([(2, 23, 0.52, 0.21, 0.21)])
        frame.append([(2, 24, 0.52, 0.21, 0.21)])

    RUN=3
    if RUN==1:
        # Filter các bboxes, khác nhau về nhãn và số lượng đều được, nhưng không fileter độ dài window theo từng nhãn
        mFilter=fnRemove_outliers_online()
        for cnt,frame1 in enumerate(frame):
            print(cnt, frame1 )
            vl=mFilter(frame1)
            print(cnt, vl )
            print()

    if RUN==2:
        mLabelFilter = clsLabelFilter_list_bboxes(3)

        for cnt,frame1 in enumerate(frame):
            vl = mLabelFilter.process_value(frame1)
            print(cnt, frame1 )
            print(cnt, vl )
            print()
    if RUN==3:
        class Param:
            def __init__(self, ):
                self.ClassNames = {
                    0: "Blue Hat",
                    1: "W Hat",
                    2: "Black Hat",
                    3: "Person",
                    4: "Blue",
                    5: "White",
                    6: "Black",
                    7: "No Shirt",
                    8: "Bend",
                }
                self.remove_outliers_online_windows_len_by_class_name = {
                    "default": 3,
                    0: 2,
                    1: 2,
                    2: 4,
                    3: 2,
                    4: 2,
                    5: 2,
                    6: 2,
                    7: 2,
                    8: 1,
                }
        mParams=Param()
        # Filter các bboxes, khác nhau về nhãn và số lượng đều được, có fileter độ dài window theo từng nhãn
        mFilter_Frames = clsFilter_Frames_sep_by_labels(ClassNames=mParams.ClassNames, windows_len_by_class=mParams.remove_outliers_online_windows_len_by_class_name)
        for cnt,frame1 in enumerate(frame):
            print(cnt, ">", frame1)
            bboxes_out=mFilter_Frames.Update(frame1, stt=cnt)
            print(cnt,'-', bboxes_out)
            print()
